/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.

   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.

   This program comes with absolutely no warranty.

   gfx.cpp - Graphics routines.
*/

/* the includes */
#include <math.h>
#include "deu.h"

#include "zethdat.h"

extern "C" {

extern DATAFILE *zethdat;

BYTE DPal[768];
int Palette[PALETTE_SIZE];

BITMAP *icon_bmp;

FONT *DethFont;
FONT *SmallFont;


/* the global variables */
double OrigX;			/* the X origin */
double OrigY;			/* the Y origin */
double Scale;			/* the scale value */
int PointerX;			/* X position of pointer */
int PointerY;			/* Y position of pointer */
int ScrMaxX;			/* maximum X screen coord */
int ScrMaxY;			/* maximum Y screen coord */
int ScrCenterX;			/* X coord of screen center */
int ScrCenterY;			/* Y coord of screen center */
int Fullscreen;			/* Fullscreen or not */

}	// extern "C"

static int current_color;
static bool kb_inited = false;

// [CR] To make a wide line with Allegro, we draw a perpendicular line
//      for every point along the normal line
static int xoff, yoff;
static void wide_line(BITMAP *bmp, int x, int y, int d)
{
	line(bmp, x-xoff, y-yoff, x+xoff, y+yoff, d);
}

static void dotted_line(BITMAP *bmp, int x, int y, int d)
{
	if(!((x+y)&3))
		putpixel(bmp, x, y, d);
}

static const int res[][2] = {{0,0},{0,0},{640,480},{800,600},
							 {1024,768},{1280,960},{1600,1200}};


CGraphics *video;

CGraphics::CGraphics()
{
}

CGraphics::CGraphics(int mode)
	: line_dotted(false), line_width(1)
{
	int i;

	// [CR] Try to use the desktop color depth(it's more reliable, and
	//      faster when windowed)
	int depths[] = { desktop_color_depth(), 32, 24, 16, 15, 8, 0 };
	if(depths[0] <= 0)
	{
		for(i = 0;depths[i];++i)
			depths[i] = depths[i+1];
	}

	int card = (Fullscreen ? GFX_AUTODETECT_FULLSCREEN :
	                         GFX_AUTODETECT_WINDOWED);

	for(i = 0;depths[i];++i)
	{
		set_color_depth(depths[i]);
		if(set_gfx_mode(card, res[mode][0], res[mode][1], 0, 0) == 0)
			break;
	}

	// If we couldn't set a mode, try to toggle Fullscreen
	if(depths[i] == 0)
	{
		Fullscreen = !Fullscreen;
		card = (Fullscreen ? GFX_AUTODETECT_FULLSCREEN :
	                         GFX_AUTODETECT_WINDOWED);

		for(i = 0;depths[i];++i)
		{
			set_color_depth(depths[i]);
			if(set_gfx_mode(card, res[mode][0], res[mode][1], 0, 0) == 0)
				break;
		}

		if(depths[i] == 0)
		{
			this->~CGraphics();
			allegro_message("Couldn't set mode %ix%i!\n%s\n", res[mode][0], res[mode][1], allegro_error);
			return;
		}
	}

	if(!kb_inited)
	{
		if(install_keyboard() != 0)
		{
			this->~CGraphics();
			allegro_message("Error installing keyboard handler!\n");
			return;
		}
		kb_inited = true;
	}

	if(set_display_switch_mode(SWITCH_BACKGROUND) == -1)
		set_display_switch_mode(SWITCH_BACKAMNESIA);

	SetDoomPalette(0);

	clear_bitmap(screen);

	buffer = create_bitmap(SCREEN_W, SCREEN_H);
	if(!buffer)
	{
		this->~CGraphics();
		allegro_message("Error creating offscreen surface!\n");
		return;
	}
	clear_to_color(buffer, bitmap_mask_color(buffer));

	icon_bmp = create_bitmap(32, 32);
	if(!icon_bmp)
	{
		this->~CGraphics();
		allegro_message("Error creating icon buffer!\n");
		return;
	}

    ScrMaxX = SCREEN_W;
    ScrMaxY = SCREEN_H;
    ScrCenterX = ScrMaxX / 2;
    ScrCenterY = ScrMaxY / 2;

	DethFont = (FONT*)zethdat[DETHFONT].dat;
	SmallFont = (FONT*)zethdat[SMALLFNT].dat;

	gui_fg_color = TranslateToGameColor(BLACK);
	gui_mg_color = TranslateToGameColor(DARKGRAY);
	gui_bg_color = TranslateToGameColor(WHITE);

	current_buffer = screen;
}

CGraphics::~CGraphics()
{
	if(kb_inited)
	{	
		remove_keyboard();
		kb_inited = false;
	}

	if(buffer)
		destroy_bitmap(buffer);
	buffer = NULL;
	if(icon_bmp)
		destroy_bitmap(icon_bmp);
	icon_bmp = NULL;

	set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
}


void CGraphics::update(int x, int y, int x_end, int y_end)
{
	scare_mouse_area(x, y, x_end - x, y_end - y);
	masked_blit(buffer, screen, x, y, x, y, x_end - x, y_end - y);
	unscare_mouse();

	clear_to_color(buffer, bitmap_mask_color(buffer));
}


void CGraphics::clear_map(int top, int maxy)
{
	HideMousePointer();
	rectfill(current_buffer, 0, top, ScrMaxX-1, ScrMaxY - maxy, TranslateToGameColor(BLACK));
	ShowMousePointer();
}

void CGraphics::set_write_mode(int mode)
{
	switch (mode)
	{
		case COPY_PUT:
			drawing_mode(DRAW_MODE_SOLID, NULL, 0, 0);
			break;
		case XOR_PUT:
			drawing_mode(DRAW_MODE_XOR, NULL, 0, 0);
			break;
	}
}


void CGraphics::draw_text(int x, int y, int color, int shad_color, int center, bool small, const char *msg, ...)
{
    static int lastX;
    static int lastY;
    char temp[256];
	va_list args;
	int len;
	FONT *f;

    if (x == -1)
		x = lastX;
    if (y == -1)
		y = lastY + 10;

	lastX = x;
	lastY = y;

	va_start(args, msg);
	len = vsprintf(temp, msg, args);
	va_end(args);

	f = (small ? SmallFont : DethFont);

	if(center)
		x -= text_length(f, temp) >> 1;

	HideMousePointer();
	if(shad_color != -1)
		textout_ex(current_buffer, f, temp, x+1, y+1, shad_color, -1);

	textout_ex(current_buffer, f, temp, x, y, color, -1);
	ShowMousePointer();
}

void CGraphics::draw_line(int x1, int y1, int x2, int y2, int color)
{
	if(x1 <= x2)
	{
		if(x1 >= ScrMaxX || x2 < 0)
			return;
	}
	else
	{
		if(x1 < 0 || x2 >= ScrMaxX)
			return;
	}

	if(y1 <= y2)
	{
		if(y1 >= ScrMaxY || y2 < 0)
			return;
	}
	else
	{
		if(y1 < 0 || y2 >= ScrMaxY)
			return;
	}

	HideMousePointer();
	if(line_width > 1)
	{
		int dx = x2 - x1;
		int dy = y2 - y1;

		// Special cases
		if(!dx)
		{
			int fat = line_width / 2;
			rectfill(current_buffer, x1-fat, y1, x2+fat, y2, color);
		}
		else if(!dy)
		{
			int fat = line_width / 2;
			rectfill(current_buffer, x1, y1-fat, x2, y2+fat, color);
		}
		else
		{
			// Sus out length
			int length = int(hypot(dy, dx)+0.5);

			// [CR] This seems to work better
			xoff = line_width * -dy / length;
			yoff = line_width *  dx / length;

			if(ABS(dx-dy) <= 4)
				xoff -= 1;

			do_line(current_buffer, x1, y1, x2, y2, color, wide_line);
		}
	}
	else if(line_dotted)
		do_line(current_buffer, x1, y1, x2, y2, color, dotted_line);
	else
		line(current_buffer, x1, y1, x2, y2, color);
	ShowMousePointer();
}

void CGraphics::draw_map_circle(int mapXcenter, int mapYcenter, int mapRadius, int color)
{
	int x = SCREENX(mapXcenter);
	int y = SCREENY(mapYcenter);
	int r = (int)(mapRadius * Scale);

	if(x-r >= ScrMaxX || x+r < 0 || y-r >= ScrMaxY || y+r < 0)
		return;

	HideMousePointer();
	circle(current_buffer, x, y, r, color);
	ShowMousePointer();
}

void CGraphics::draw_map_box(int x1, int y1, int x2, int y2, int color)
{
	HideMousePointer();
	rect(current_buffer, SCREENX(x1), SCREENY(y1), SCREENX(x2), SCREENY(y2), color);
	ShowMousePointer();
}

/* draw a filled in box on the screen from screen coords */
void CGraphics::draw_box(int Xstart, int Ystart, int Xend, int Yend, int color)
{
	HideMousePointer();
	rectfill(current_buffer, Xstart, Ystart, Xend, Yend, color);
	ShowMousePointer();
}

void CGraphics::draw_frame(int Xstart, int Ystart, int Xend, int Yend, int color)
{
	HideMousePointer();
	rect(current_buffer, Xstart, Ystart, Xend, Yend, color);
	ShowMousePointer();
}


void CGraphics::draw_map_vector(int mapXstart, int mapYstart, int mapXend, int mapYend, int color)
{
	int  scrXstart = SCREENX(mapXstart);
	int  scrYstart = SCREENY(mapYstart);
	int  scrXend   = SCREENX(mapXend);
	int  scrYend   = SCREENY(mapYend);
	double r       = hypot((double)(scrXstart - scrXend), (double)(scrYstart - scrYend));
	int  scrXoff   = (r >= 1.0) ? (int)((scrXstart - scrXend) * 4.0 / r * Scale) : 0;
	int  scrYoff   = (r >= 1.0) ? (int)((scrYstart - scrYend) * 4.0 / r * Scale) : 0;

	HideMousePointer();
	draw_line(scrXstart, scrYstart, scrXend, scrYend, color);

	scrXstart = scrXend + 2 * scrXoff;
	scrYstart = scrYend + 2 * scrYoff;
	draw_line(scrXstart - scrYoff, scrYstart + scrXoff, scrXend, scrYend, color);
	draw_line(scrXstart + scrYoff, scrYstart - scrXoff, scrXend, scrYend, color);
	ShowMousePointer();
}



/* draw an arrow on the screen from map coords and angle (0 - 65535) */
void CGraphics::draw_map_arrow(int mapXstart, int mapYstart, int angle, int color)
{
	int  mapXend   = mapXstart + (int)(50 * cos(angle / 10430.37835));
	int  mapYend   = mapYstart + (int)(50 * sin(angle / 10430.37835));
	int  scrXstart = SCREENX(mapXstart);
	int  scrYstart = SCREENY(mapYstart);
	int  scrXend   = SCREENX(mapXend);
	int  scrYend   = SCREENY(mapYend);
	double r       = hypot(scrXstart - scrXend, scrYstart - scrYend);
	int  scrXoff   = (r >= 1.0) ? (int)((scrXstart - scrXend) * 4.0 / r * Scale) : 0;
	int  scrYoff   = (r >= 1.0) ? (int)((scrYstart - scrYend) * 4.0 / r * Scale) : 0;

	HideMousePointer();
	draw_line(scrXstart, scrYstart, scrXend, scrYend, color);
	scrXstart = scrXend + 2 * scrXoff;
	scrYstart = scrYend + 2 * scrYoff;
	draw_line(scrXstart - scrYoff, scrYstart + scrXoff, scrXend, scrYend, color);
	draw_line(scrXstart + scrYoff, scrYstart - scrXoff, scrXend, scrYend, color);
	ShowMousePointer();
}


/* draw a filled-in 3D-box on the screen from screen coords */
void CGraphics::draw_box_3d(int Xstart, int Ystart, int Xend, int Yend)
{
	int c;

	HideMousePointer();
	c = TranslateToGameColor(DARKERGRAY);
    line(current_buffer, Xstart, Yend, Xend, Yend, c);
    line(current_buffer, Xend, Ystart, Xend, Yend, c);
    c = TranslateToGameColor(LIGHTGRAY);
    line(current_buffer, Xstart, Ystart, Xend, Ystart, c);
    line(current_buffer, Xstart, Ystart, Xstart, Yend, c);
	rectfill(current_buffer, Xstart + 1, Ystart + 1, Xend - 1, Yend - 1,
	         TranslateToGameColor(DARKGRAY));
	ShowMousePointer();
}


/* draw a hollow 3D-box on the screen from screen coords */
void CGraphics::draw_box_hollow(int Xstart, int Ystart, int Xend, int Yend)
{
	int c;

	HideMousePointer();
	c = TranslateToGameColor(LIGHTGRAY);
    line(current_buffer, Xstart, Yend, Xend, Yend, c);
    line(current_buffer, Xend, Ystart, Xend, Yend, c);
    c = TranslateToGameColor(DARKERGRAY);
    line(current_buffer, Xstart, Ystart, Xend, Ystart, c);
    line(current_buffer, Xstart, Ystart, Xstart, Yend, c);
	rectfill(current_buffer, Xstart + 1, Ystart + 1, Xend - 1, Yend - 1,
	         TranslateToGameColor(BLACK));
	ShowMousePointer();
}

void CGraphics::draw_bar(int Xstart, int Ystart, int Width)
{
	HideMousePointer();
    line(current_buffer, Xstart, Ystart, Xstart + Width - 1, Ystart,
	     TranslateToGameColor(DARKERGRAY));
    line(current_buffer, Xstart, Ystart + 1, Xstart + Width, Ystart + 1,
	     TranslateToGameColor(LIGHTGRAY));
	ShowMousePointer();
}

// [RH] Why did the buttons have an illegible bright shadow?
void CGraphics::draw_button(int Xstart, int Ystart, int Width, int Height, const char *Text, bool active)
{
	int c;

	HideMousePointer();
	rect(current_buffer, Xstart - 1, Ystart - 1, Xstart + Width + 1, Ystart + Height + 1, TranslateToGameColor(BLACK));

	c = TranslateToGameColor(LIGHTGRAY);
	line(current_buffer, Xstart, Ystart, Xstart + Width, Ystart, c);
	line(current_buffer, Xstart, Ystart, Xstart, Ystart + Height, c); 
    c = TranslateToGameColor(DARKERGRAY);
    line(current_buffer, Xstart, Ystart + Height, Xstart + Width, Ystart + Height, c);
    line(current_buffer, Xstart + Width, Ystart, Xstart + Width, Ystart + Height, c);

	rectfill(current_buffer, Xstart + 1, Ystart + 1, Xstart + Width - 1, Ystart + Height - 1, TranslateToGameColor(DARKGRAY));
	textout_centre_ex(current_buffer, SmallFont, Text, Xstart + (Width >> 1), Ystart + (Height >> 1) - 6 + 1,
	                  TranslateToGameColor(active ? BLACK : DARKERGRAY), -1);
	ShowMousePointer();
}

void CGraphics::draw_button_in(int Xstart, int Ystart, int Width, int Height, char *Text)
{
	int c;

	HideMousePointer();
	c = TranslateToGameColor(LIGHTGRAY);
	line(current_buffer, Xstart, Ystart + Height, Xstart + Width, Ystart + Height, c);
	line(current_buffer, Xstart + Width, Ystart, Xstart + Width, Ystart + Height, c); 
	c = TranslateToGameColor(DARKERGRAY);
	line(current_buffer, Xstart, Ystart, Xstart + Width, Ystart, c);
	line(current_buffer, Xstart, Ystart, Xstart, Ystart + Height, c); 

	rectfill(current_buffer, Xstart + 1, Ystart + 1, Xstart + Width - 1, Ystart + Height - 1, TranslateToGameColor(DARKGRAY));
	textout_centre_ex(current_buffer, SmallFont, Text, Xstart + (Width >> 1), Ystart + (Height >> 1) - 6 + 1,
	                  TranslateToGameColor(BLACK), -1);
	ShowMousePointer();
}

void CGraphics::draw_highlight(int Xstart, int Ystart, int Width, int Height)
{
	int c;

	HideMousePointer();
	c = TranslateToGameColor(DARKERGRAY);
	line(current_buffer, Xstart, Ystart + Height, Xstart + Width, Ystart + Height, c);
	line(current_buffer, Xstart + Width, Ystart, Xstart + Width, Ystart + Height, c);
	c = TranslateToGameColor(LIGHTGRAY);
	line(current_buffer, Xstart, Ystart, Xstart + Width, Ystart, c);
	line(current_buffer, Xstart, Ystart, Xstart, Ystart + Height, c);

	rectfill(current_buffer, Xstart + 1, Ystart + 1, Xstart + Width - 1,
	         Ystart + Height - 1, TranslateToGameColor(BLUE));
	ShowMousePointer();
}


void CGraphics::draw_window(int x, int y, int width, int height, char *title, int color)
{
	int dx = x + width;
	int dy = y + height;
	int c;

	HideMousePointer();
	c = TranslateToGameColor(LIGHTGRAY);
	line(current_buffer, x, y, dx - 1, y, current_color);
	line(current_buffer, x, y, x, dy - 1, current_color);
	c = TranslateToGameColor(DARKERGRAY);
	line(current_buffer, x, dy, dx , dy, current_color);
	line(current_buffer, dx, y, dx, dy, current_color);

	rectfill(current_buffer, x + 1, y + 1, dx - 1, dy - 1, TranslateToGameColor(DARKGRAY));
	rectfill(current_buffer, x + 3, y + 3, dx - 3, y + 17, TranslateToGameColor(color));

	textout_ex(current_buffer, SmallFont, title, x+6, y+4, TranslateToGameColor(BLACK), -1);
	ShowMousePointer();
}

void CGraphics::highlight_area(int x, int y, int width, int height, const char *text, int color, int textcolor)
{
	int dx = x + width;
	int dy = y + height;

	HideMousePointer();
	rectfill(current_buffer, x, y, dx - 1, dy - 1, TranslateToGameColor(color));
	textout_ex(current_buffer, DethFont, text, x+3, y+2, TranslateToGameColor(textcolor), -1);
	ShowMousePointer();
}

void CGraphics::box_area(int x, int y, int width, int height, int color)
{
	int dx = x + width;
	int dy = y + height;

	HideMousePointer();
	rect(current_buffer, x, y, dx, dy, TranslateToGameColor(color));
	ShowMousePointer();
}

void CGraphics::draw_inset(int Xstart, int Ystart, int Width, int Height)
{
	int c;

	HideMousePointer();
	c = TranslateToGameColor(LIGHTGRAY);
	line(current_buffer, Xstart, Ystart + Height, Xstart + Width, Ystart + Height, c);
	line(current_buffer, Xstart + Width, Ystart, Xstart + Width, Ystart + Height, c); 
	c = TranslateToGameColor(DARKERGRAY);
	line(current_buffer, Xstart, Ystart, Xstart + Width, Ystart, c);
	line(current_buffer, Xstart, Ystart, Xstart, Ystart + Height, c); 
	ShowMousePointer();
}


/***************
 [CR] Begin C compatibility functions
 ***************/
extern "C" {

/* initialise the graphics display */
void InitGfx(void)
{
	if (VideoMode < 2 || VideoMode > 6)
		VideoMode = 2;

	if(gfx_driver)
	{
		if(ScrMaxX == res[VideoMode][0] && ScrMaxY == res[VideoMode][1])
		{
			if((!gfx_driver->windowed && Fullscreen) ||
			   (gfx_driver->windowed && !Fullscreen))
				return;
		}
	}

	delete video;
	video = new CGraphics(VideoMode);
	if(!gfx_driver || !video)
	{
		delete video;
		video = NULL;

		exit(1);
	}

	CheckMouseDriver();
}

/* terminate the graphics display */
void TermGfx(void)
{
	// [CR] We don't want to physically quit from a graphics mode if
	//      there's no console to fall to
#ifdef ALLEGRO_CONSOLE_OK
	if(gfx_driver && !gfx_driver->windowed)
	{	
		delete video;
		video = NULL;
	}
#endif
}

BITMAP *GetBuffer(void)
{
	return video->get_buffer();
}

void BufferDrawing(Bool set)
{
	video->buffer_drawing(!!set);
}

void ScreenUpdate(int x, int y, int x_end, int y_end)
{
	video->update(x, y, x_end, y_end);
}

/* clear the map screen */
/*AJB*/
void ClearMapScreen(int top, int maxy)
{
	video->clear_map(top, maxy);
}

/* set the current drawing color */
void SetColor(int color)
{
	current_color = TranslateToGameColor(color);
}

/*
	set the write mode
 */
void SetWriteMode(int mode)
{
	video->set_write_mode(mode);
}

void SetLineDotted(Bool dotted)
{
	video->set_line_dotted(!!dotted);
}

void SetLineThickness(int width)
{
	video->set_line_thickness(width);
}


/* draw a line on the screen from screen coords */
void DrawScreenLine(int Xstart, int Ystart, int Xend, int Yend)
{
	video->draw_line(Xstart, Ystart, Xend, Yend, current_color);
}


/* draw a line on the screen from map coords */
void DrawMapLine(int mapXstart, int mapYstart, int mapXend, int mapYend)
{
	video->draw_line(SCREENX(mapXstart), SCREENY(mapYstart),
	                 SCREENX(mapXend), SCREENY(mapYend), current_color);
}



/* draw a circle on the screen from map coords */
void DrawMapCircle(int mapXcenter, int mapYcenter, int mapRadius)
{
	video->draw_map_circle(mapXcenter, mapYcenter, mapRadius, current_color);
}


/* draw a hollow box on the map */
void DrawMapHollowBox(int x1, int y1, int x2, int y2)
{
	video->draw_map_box(x1, y1, x2, y2, current_color);
}

/* draw a filled in box on the screen from screen coords */
void DrawScreenBox(int Xstart, int Ystart, int Xend, int Yend)
{
	video->draw_box(Xstart, Ystart, Xend, Yend, current_color);
}

void DrawScreenFrame(int Xstart, int Ystart, int Xend, int Yend)
{
	video->draw_frame(Xstart, Ystart, Xend, Yend, current_color);
}

/* draw an arrow on the screen from map coords */
void DrawMapVector(int mapXstart, int mapYstart, int mapXend, int mapYend)
{
	video->draw_map_vector(mapXstart, mapYstart, mapXend, mapYend, current_color);
}

/* draw an arrow on the screen from map coords and angle (0 - 65535) */
void DrawMapArrow(int mapXstart, int mapYstart, UBCINT angle)
{
	video->draw_map_arrow(mapXstart, mapYstart, angle, current_color);
}

/* draw a filled-in 3D-box on the screen from screen coords */
void DrawScreenBox3D(int Xstart, int Ystart, int Xend, int Yend)
{
	video->draw_box_3d(Xstart, Ystart, Xend, Yend);
}

/* draw a hollow 3D-box on the screen from screen coords */
void DrawScreenBoxHollow(int Xstart, int Ystart, int Xend, int Yend)
{
	video->draw_box_hollow(Xstart, Ystart, Xend, Yend);
}



void DrawScreenBar(int Xstart, int Ystart, int Width)
{
	video->draw_bar(Xstart, Ystart, Width);
}

// [RH] Why did the buttons have an illegible bright shadow?
void DrawScreenButton( int Xstart, int Ystart, int Width, int Height, char *Text)
{
	video->draw_button(Xstart, Ystart, Width, Height, Text, true);
}

void DrawScreenButtonGray(int Xstart, int Ystart, int Width, int Height, char *Text)
{
	video->draw_button(Xstart, Ystart, Width, Height, Text, false);
}

void DrawScreenButtonIn(int Xstart, int Ystart, int Width, int Height, char *Text)
{
	video->draw_button_in(Xstart, Ystart, Width, Height, Text);
}

void DrawScreenButtonOut(int Xstart, int Ystart, int Width, int Height)
{
	video->draw_highlight(Xstart, Ystart, Width, Height);
}


void DrawScreenWindow(int x, int y, int width, int height, char *title, int color)
{
	video->draw_window(x, y, width, height, title, color);
}

void ScreenHighlightArea(int x, int y, int width, int height, char *text, int color, int textcolor)
{
	video->highlight_area(x, y, width, height, text, color, textcolor);
}

void ScreenBoxArea(int x, int y, int width, int height, int color)
{
	video->box_area(x, y, width, height, color);
}

void DrawScreenInset(int Xstart, int Ystart, int Width, int Height)
{
	video->draw_inset(Xstart, Ystart, Width, Height);
}

/*
   write text to the screen
*/
void DrawScreenText(int Xstart, int Ystart, char *msg, ...)
{
    char temp[256] = "";
    va_list args;
	int i;

    va_start( args, msg);
	vsprintf( temp, msg, args);
    va_end( args);

    for(i = 0;temp[i];++i)
    {
        if(temp[i] == '%')
		{
			memmove(temp+i+1, temp+i, sizeof(temp)-i-1);
			++i;
		}
    }

	video->draw_text(Xstart, Ystart, current_color, -1, false, false, temp);
}

void DrawScreenTextSmall(int Xstart, int Ystart, char *msg, ...)
{
    char temp[256] = "";
    va_list args;
	int i;

    va_start( args, msg);
	vsprintf( temp, msg, args);
    va_end( args);

    for(i = 0;temp[i];++i)
    {
        if(temp[i] == '%')
		{
			memmove(temp+i+1, temp+i, sizeof(temp)-i-1);
			++i;
		}
    }

	video->draw_text(Xstart, Ystart, current_color, -1, false, true, temp);
}

void DrawScreen3DText(int Xstart, int Ystart, int col, int scol, char *msg, ...)
{
    char temp[256] = "";
    va_list args;
	int i;

    va_start( args, msg);
	vsprintf( temp, msg, args);
    va_end( args);

    for(i = 0;temp[i];++i)
    {
        if(temp[i] == '%')
		{
			memmove(temp+i+1, temp+i, sizeof(temp)-i-1);
			++i;
		}
    }

	video->draw_text(Xstart, Ystart, TranslateToGameColor(col), TranslateToGameColor(scol), false, false, temp);
}

/*
   load one "playpal" palette and change all palette colours
   */
void SetDoomPalette(int playpalnum)
{
	PALETTE pal;
	BYTE	gamma[256];
	MDirPtr	dir, next;
	int	n;

	dir = FindMasterDir(MasterDir, "PLAYPAL");
	if(!dir || playpalnum < 0 || playpalnum > 13)
		return;

	while(dir->next && (next = FindMasterDir(dir->next, "PLAYPAL")) != NULL)
		dir = next;

	BasicWadSeek( dir->wadfile, dir->dir.start);
	for (n = 0; n <= playpalnum; n++)
		BasicWadRead( dir->wadfile, DPal, 768L);

	if (Gamma > 0 && Gamma <= 4) 
	{
		static const double gmult[5] = {1.0, 0.75, 0.55, 0.4, 0.3};
		double mult = gmult[Gamma];

		for (n = 0; n < 256; ++n)
		{
			gamma[n] = (BYTE)(pow(((double)n / 255.0), mult) * 255.0);
		}
	}
	else
	{
		for (n = 0; n < 256; ++n)
			gamma[n] = n;
	}

	for (n = 0; n < 256; n++)
	{
		pal[n].r = gamma[DPal[3*n]] >> 2;
		pal[n].g = gamma[DPal[3*n+1]] >> 2;
		pal[n].b = gamma[DPal[3*n+2]] >> 2;
	}
	set_palette(pal);
}

/*
   translate a standard color to Doom palette 0 (approx.)
   */

int TranslateToGameColor(int color)
{
	int c = palette_color[Palette[color]];

	if(c == bitmap_mask_color(screen))
	{
		if(c == 0)
			c = makecol(0, 0, 0);
		else
			--c;
	}

	return c;
}


/*
   translate (dx, dy) into an integer angle value (0-65535)
   */

UBCINT ComputeAngle( int dx, int dy)
{
    return (UBCINT) (atan2( (double) dy, (double) dx) * 10430.37835 + 0.5);
    /* Yes, I know this function could be in another file, but */
    /* this is the only source file that includes <math.h>...  */
}



/*
   compute the distance from (0, 0) to (dx, dy)
   */

UBCINT ComputeDist( int dx, int dy)
{
    return (UBCINT) (hypot( (double) dx, (double) dy) + 0.5);
    /* Yes, I know this function could be in another file, but */
    /* this is the only source file that includes <math.h>...  */
}



/*
   insert the vertices of a new polygon
   */

void InsertPolygonVertices( int centerx, int centery, int sides, int radius)
{
    int n;
    
    for (n = 0; n < sides; n++)
	{
		InsertObject( OBJ_VERTEXES, -1,
			centerx + (int) ((double) radius * cos( 2*M_PI * (double) n / (double) sides)),
			centery + (int) ((double) radius * sin( 2*M_PI * (double) n / (double) sides)));
	}
    /* Yes, I know... etc. */
}



/*
   move (x, y) to a new position: rotate and scale around (0, 0)
*/
void RotateAndScaleCoords(double *x, double *y, double angle, double scale)
{
	double r = hypot(*x,*y);
	double theta = atan2(*y,*x);

    *x = r * scale * cos(theta + angle);
    *y = r * scale * sin(theta + angle);
}

}	// extern "C"

/* end of file */
